#wx_display_main.py 17Jan2024  crs, Author
""" Startup wxpython display
"""
import argparse
import wx

from select_trace import SlTrace
from wx_braille_display import BrailleDisplay
from wx_tk_rpc_user import TkRPCUser
from wx_braille_cell_list import BrailleCellList

if __name__ == '__main__':      # Required because we use multiprocessing
                                # in some modules e.g. pyttsx_proc.py
                                #
    SlTrace.clearFlags()
    subprocess = False
    id_title = ""       # Identification title
    title = None
    src_file = __file__   # Replaced with remote source file name
    bdlist = None
    host_port = 50040
    host_port = 50020
    port_inc = None
    parser = argparse.ArgumentParser()
    parser.add_argument('--id_title', type=str, dest='id_title', default=id_title)
    parser.add_argument('--title', type=str, dest='title', default=title)
    parser.add_argument('--src_file', type=str, dest='src_file', default=src_file)
    parser.add_argument('--bdlist', type=str, dest='bdlist', default=bdlist)
    parser.add_argument('--host_port', type=int, dest='host_port', default=host_port)
    parser.add_argument('--port_inc', type=int, dest='port_inc', default=port_inc)
    parser.add_argument('--subprocess', action='store_true', dest='subprocess', default=subprocess)
    #parser.add_argument('--ncol=', type=int, dest='ncol', default=ncol)
    args = parser.parse_args()             # or die "Illegal options"
    SlTrace.lg(f"args: {args}\n")
    id_title = args.id_title
    title = args.title
    src_file = args.src_file
    bdlist = args.bdlist
    host_port = args.host_port
    port_inc = args.port_inc
    
    #tkh = TkRPCHost()
    SlTrace.lg(f"wx_display_main.py setting up tk link", "tk_link")
    tkr = TkRPCUser(host_port=host_port)
    SlTrace.lg(f"wx_display_main.py after tk link setup", "tk_link")
    app = wx.App()
    SlTrace.lg(f"wx_display_main.py retrieve cell_specs", "tk_link")
    cells = tkr.get_cell_specs()  # gets (ix,iy,color)*
    SlTrace.lg(f"wx_display_main.py tkr.get_cell_specs() cells: {cells}", "cell_specs")
    cell_list = BrailleCellList(cells)  # converts either to BrailleCell
    bdlist = cell_list.to_string()
    bd = BrailleDisplay(tkr, id_title=id_title, title=title,
                        src_file=src_file, display_list=bdlist)
    bd.display(title=title)
    tkr.setup_from_host_requests()      # Wait till after initial display
    SlTrace.lg("wx_display_main.py after bd.display", "tk_link")
    app.MainLoop()




#wx_tk_rpc_user.py  08Feb2024  crs, split from wx_tk_rem_access.py
""" Client/User part of wx_tk_rem_host.py, wx_tk_rem_user.py pair
Supporting communication betw
"""
import socket as sk
import threading as th
import queue
import pickle
import time
import wx

from wx_rpc import RPCClient
from wx_rpc import RPCServer


from select_trace import SlTrace

        
class TkRPCUser:
    """User (remote) control requesting canvas information
    Using socket client
    """
    def __init__(self, host_name='localhost',
                 host_port=None,
                 max_recv=2**16, simulated=False,
                 cmd_time_ms= 500,
                 figure=1):
        """ Handle user (wxPython) side of communications
        :host_name: servere host name default: localhost - same machine
        :host_port: port to send server requests
        :max_recv: maximum data recieved length in bytes
                    default: 2**16
        :cmd_time_ms: maximum between cmd check in seconds
                    default: .5 sec
        :simulated: True: simulate tk input default: False
        :figure: simulated figure 1 - spokes, 2 - square
                default=1 spokes
        """
        SlTrace.lg("TkRPCUser() __init__() BEGIN", "tk_link")
        self.adw = None         # Set when ready
        self.snapshots = []     # snapshot adw, if any

        self.host_name = host_name
        if host_port is None:
            if not simulated:
                raise Exception("server port is not specified")
        else: 
            self.host_port = host_port
            self.user_port = host_port+1        # ??? better choice?
        
        self.max_recv = max_recv
        self.cmd_time_ms = cmd_time_ms
        self.simulated = simulated
        SlTrace.lg("TkRPCUser() __init__()", "tk_link")
        if simulated:
            self.make_simulated(figure=figure)
            return

        self.to_host = RPCClient(self.host_name, self.host_port)
        self.to_host.connect()

        #self.setup_from_host_requests()
        
    def setup_from_host_requests(self):           
        self.from_host_server = RPCServer(self.host_name, self.user_port)

        self.from_host_server.registerMethod(self.snapshot)

        th.Thread(target=self.cmd_in_th_proc).start()
        self.to_host.setup_calling_user()  # enable req from host
        
    def cmd_in_th_proc(self):
        SlTrace.lg(f"USER: server_host_th_proc", "tk_link")
        self.from_host_server.run()

    """
    User based functions
    remotely requested from the host  
    """
    
    def snapshot(self, title=None, snapshot_num=None):
        """ Create display snapshot
        :title: display title
        """
    
        SlTrace.lg(f"""USER: snapshot(self, title={title},
                   snapshot_num={snapshot_num})""", "snapshot")
        SlTrace.lg(f"""wx.CallAfter(self.snapshot_direct,
                   title={title}, snapshot_num={snapshot_num}""", "snapshot")
        wx.CallAfter(self.snapshot_direct, title=title,
                     snapshot_num=snapshot_num)

    def snapshot_direct(self, title=None, snapshot_num=None):
        """ Direct call, from event processor
        :title: display title
        """
        
        SlTrace.lg(f"USER: snapshot_direct(self, title={title}) direct call", "tk_link")
        if self.adw is None:
            SlTrace.lg("USER: AudioDrawWindow not set - snapshot ignored")
            return
        
        SlTrace.lg(f"USER: adw.create_audio_window(title={title},)"
                   f" snapshot_num={snapshot_num}", "tk_link")
        sn_title = f"{title}: {self.adw.title}"
        adw = self.adw.create_audio_window(title=sn_title,
                                snapshot_num=snapshot_num)
        self.snapshots.append(adw)
        #adw.Refresh()
        adw.Update()
        #adw.Show()

        self.snapshot_complete()

    def snapshot_complete(self):
        """ Signal snapshot competion
        """
        self.to_host.snapshot_complete()
                    
    def get_cell_specs(self,
                        snapshot_num=None,
                        win_fract=True, 
                        x_min=None, y_min=None,
                        x_max=None, y_max=None,
                        n_cols=None, n_rows=None):
        """ Get cell specs from remote tk canvas
        :returns: list of cell specs (ix,iy,color)
        """
        SlTrace.lg(f"""TkRPCUser:get_cell_specs(
                        snapshot_num={snapshot_num},
                        win_fract={win_fract}, 
                        x_min={x_min}, y_min={y_min},
                        x_max={x_max}, y_max={y_max},
                        n_cols={n_cols}, n_rows={n_rows})
                        """, "tk_link")        
        if self.simulated:
            return self.get_cell_specs_simulated(
                        x_min=x_min, y_min=y_min,
                        win_fract=win_fract, 
                        x_max=x_max, y_max=y_max,
                        n_cols=n_cols, n_rows=n_rows)
        
        call_num = self.to_host.get_cell_specs(
                    TK_EXECUTE_IN_MAIN_THREAD=True,
                    snapshot_num=snapshot_num,
                    x_min=x_min, y_min=y_min,
                    x_max=x_max, y_max=y_max,
                    n_cols=n_cols, n_rows=n_rows)
        SlTrace.lg(f"TkRPCUser:get_cell_specs call_num:{call_num}", "tk_link")
        while True:
            ret = self.to_host.get_ret(call_num)
            if ret is not None:
                break
        SlTrace.lg(f"TkRPCUser:get_cell_specs[{call_num}] ret:{ret}", "tk_link")
        return ret    
    
    def get_cell_specs_simulated(self, 
                        win_fract=True, 
                        x_min=None, y_min=None,
                        x_max=None, y_max=None,
                        n_cols=None, n_rows=None):
        """ Get cell specs from remote tk canvas
        :returns: list of cell specs (ix,iy,color)
        """        
        return self.sim_cg.get_cell_specs(
                        win_fract=win_fract, 
                        x_min=x_min, y_min=y_min,
                        x_max=x_max, y_max=y_max,
                    n_cols=n_cols, n_rows=n_rows)


    def is_inbounds_ixy(self, *ixy):
        """ Check if ixy pair is in bounds
        :ixy: if tuple ix,iy pair default: current location
              else ix,iy indexes
            ix: cell x index default current location
            iy: cell y index default current location
        :returns: True iff in bounds else False
        """
        if self.simulated:
            return self.is_inbounds_ixy_simulated(self, *ixy)
        
        return self.to_host.is_inbounds_ixy(*ixy)

    def is_inbounds_ixy_simulated(self, *ixy):
        """ Check if ixy pair is in bounds
        :ixy: if tuple ix,iy pair default: current location
              else ix,iy indexes
            ix: cell x index default current location
            iy: cell y index default current location
        :returns: True iff in bounds else False
        """
    

    def get_cell_rect_tur(self, ix, iy):
        """ Get cell's turtle rectangle x, y  upper left, x,  y lower right
        :ix: cell x index
        :iy: cell's  y index
        :returns: (min_x,max_y, max_x,min_y)
        """
        if self.simulated:
            return self.get_cell_rect_tur_simulated(
                        ix=ix,iy=iy)
        
        call_num = self.to_host.get_cell_rect_tur(ix, iy,
                    TK_EXECUTE_IN_MAIN_THREAD=True)
        while True:
            ret = self.to_host.get_ret(call_num)
            if ret is not None:
                break
        SlTrace.lg(f"get_cell_rect_tur ret:{ret}", "rpc,tk_link")
        return ret    
    
    def get_cell_rect_tur_simulated(self, 
                        ix,iy):
        """ Get cell rectangle
        :returns: )
        """        
        return self.sim_cg.get_cell_rect_tur(
                        ix,iy)
    
    
    def get_canvas_lims(self, win_fract=True):
        """ Get canvas limits
        :win_fract: True - fractional 0. to 1.
                    False - window coordinates
        :returns: internal (xmin, xmax, ymin, ymax)
        """
        if self.simulated:
            return self.get_canvas_lims_simulated(win_fract=win_fract)
        
        if win_fract:
            return (0.,1., 0., 1.)
        
        #return self.to_host.get_canvas_lims(win_fract)
        
        call_num = self.to_host.get_canvas_lims(win_fract,
                    TK_EXECUTE_IN_MAIN_THREAD=True)
        while True:
            ret = self.to_host.get_canvas_lims(call_num)
            if ret is not None:
                break
        SlTrace.lg(f"get_canvas_lims ret:{ret}", "tk_link")
        return ret    
                        
    
    def get_canvas_lims_simulated(self, win_fract=True):
        """ Simulate access to canvas info
        :returns: internal (xmin, xmax, ymin, ymax)
        """
        if win_fract:
            return (0.,1.,0.,1.)
        
        winfo_width = self.sc_width
        winfo_height = self.sc_height
        xmax = winfo_width/2
        xmin = -xmax
        ymax = winfo_height/2
        ymin = -ymax
        return (xmin,xmax,ymin,ymax)

    
    def test_command(self, 
                        message="Test message"):
        """ Do test message
        :message: test message default:"Test message"
        :returns: "Answer: <message>
        """
        return self.to_host.test_command(message)

    def make_simulated(self, figure=1):
        """ Setup local tkinter canvas + canvas_grid
        to provide direct access for local calls
            get_canvas_specs
            get_canvas_lims
        :figure: simulated tk figure
                0 - empty 
                1 - spokes ( like wx_spokes.py)
                2 - square ( like wx_square_colors.py)
        """
        if figure == 0:
            self.make_simulated_empty()
        elif figure  == 1:
            self.make_simulated_spokes()
        elif figure == 2:
            self.make_simulated_square()
        else:
            raise Exception(f"Unsupported figure: {figure}")

    def make_simulated_spokes(self):
        """ simulate wx_spokes.py
        """
        import turtle as tur
        from wx_canvas_grid import CanvasGrid
        sc_width = 800
        sc_height = 900
        self.sc_width = sc_width
        self.sc_height = sc_height
        screen = tur.Screen()
        screen.setup(sc_width, sc_height)
        tur.speed("fastest")
        colors = ["red","orange","yellow","green",
                  "blue","indigo","violet"]
        sp_len = 250
        dot_size = sp_len/3
        for colr in colors:
            tur.color(colr)
            tur.forward(sp_len)
            tur.dot(100)
            tur.backward(sp_len)
            tur.right(360/len(colors))
        canvas = tur.getcanvas()
        self.sim_cg = CanvasGrid(base=canvas)

    def make_simulated_empty(self):
        """ simulate empty screen
        """
        import turtle as tur
        from wx_canvas_grid import CanvasGrid
        sc_width = 800
        sc_height = 900
        self.sc_width = sc_width
        self.sc_height = sc_height
        screen = tur.Screen()
        screen.setup(sc_width, sc_height)
        canvas = tur.getcanvas()
        self.sim_cg = CanvasGrid(base=canvas)

    def make_simulated_square(self):
        """ simulate wx_square_colors.py
        """
        import turtle as tur
        from wx_canvas_grid import CanvasGrid
        sc_width = 800
        sc_height = 900
        self.sc_width = sc_width
        self.sc_height = sc_height
        screen = tur.Screen()
        screen.setup(sc_width, sc_height)
        tur.speed("fastest")
        colors = ["red","orange","yellow","green"]

        for colr in colors:
            tur.width(40)
            tur.color(colr)
            tur.forward(200)
            tur.right(90)
        canvas = tur.getcanvas()
        self.sim_cg = CanvasGrid(base=canvas)

    def set_adw(self, adw):
        """ Setup AudioDrawWindow, to support snapshot
            only first instance
        :adw: AudioWindow reference
        """
        SlTrace.lg(f"set_adw:{adw}", "adw")
        if self.adw is None:
            SlTrace.lg(f"set_adw({adw} first)", "tk_link")
            self.adw = adw
        
if __name__ == '__main__':
    import sys
    
    tkr = TkRPCUser(simulated=True)
    for msg in ["t1","t2","t3"]:
        SlTrace.lg(f"test_command({msg})")
        res = tkr.test_command(msg)
        SlTrace.lg(f"result: {res}")
    
    if len(sys.argv)  > 1:
        res = tkr.get_cell_specs()
        SlTrace.lg(f"cell_specs: {res}")


#add_src_dir.py
""" Add this file's source directory
to sys.path, if not already in sys.path
This module was created as a workaround to the
lack of our pypi installation site directory path
enabling visibility to our packages modules
This module expects to reside in our package's
site directory.  If that directory is not
found in sys.path the directory is added to sys.path
"""
import sys
import os
print(f"Initial sys.path:\n    {"\n    ".join(sys.path)}")

src_file = os.path.abspath(__file__)
print(f"add_src_dir file: {src_file}")
src_path = os.path.dirname(src_file)
if src_path not in sys.path:
    print(f"Inserting source directory {src_path}")
    sys.path.insert(0, src_path)
    print(f"sys.path:\n    {"\n    ".join(sys.path)}")

if __name__ == '__main__':

    import add_src_dir
    print(f"Selftest for {__file__}") 


# format_exception.py    03Mar2023  crs, from Stackoverflow

"""
exception string
Example:
    from format_exception import format_exception
    
    
    try:
    except AssertionError as err_msg:
        SlTrace.lg(f"Test {test_desc} FAILED: {err_msg}")
    except Exception as e:
        SlTrace.lg(f"Unexpected exception: {e}")
        SlTrace.lg("Printing the full traceback as if we had not caught it here...")
        SlTrace.lg(format_exception(e))
"""

import os
import traceback
import sys

def format_exception(e):
    exception_list = traceback.format_stack()
    exception_list = exception_list[:-2]
    exception_list.extend(traceback.format_tb(sys.exc_info()[2]))
    exception_list.extend(traceback.format_exception_only(sys.exc_info()[0], sys.exc_info()[1]))

    exception_str = "Traceback (most recent call last):\n"
    exception_str += "".join(exception_list)
    # Removing the last \n
    exception_str = exception_str[:-1]

    return exception_str

    


# show_square_loop_colors_grid.py  # standard
# Display a square with colored sides

from turtle import *		     # Get standard stuff

colors = ["red","orange","yellow","green"]

for colr in colors:
    width(40)
    color(colr)
    forward(200)
    right(90)

# Do grid to demonstrate dividing up screen
# into regions for text map for Braille
speed(0)
fudge = 40
scr_width = window_width()-fudge
x_boarder = 30
y_boarder = 25
scr_height = window_height()-fudge
x_start = -scr_width/2 + x_boarder
x_end = x_start + scr_width - x_boarder
x_len = x_end - x_start
y_start = -scr_height/2 + y_boarder
y_end = y_start + scr_height - y_boarder
y_len = y_end-y_start
ncol = 40
nrow = 25
width(1)
color("black")
# Horizontal lines
for n in range(1, nrow+1+1):
    penup()
    y_val = y_start + (n-1)*y_len/nrow
    if n < nrow+1:
        text_row_x = x_start - 40
        text_row_y = y_val + .4*y_len/nrow
        goto(x=text_row_x, y=text_row_y)
        write(f"row {n:2}") 
    goto(x=x_start, y=y_val)
    pendown()
    goto(x=x_end, y=y_val)

# Vertical lines    
for n in range(1, ncol+1+1):
    penup()
    x_val = x_start + (n-1)*x_len/ncol
    if n < ncol+1:
        text_row_x = x_val + 0.2*x_len/nrow
        text_row_y = y_start - 20
        goto(x=text_row_x, y=text_row_y)
        write(f"{n:2}") 
    goto(x=x_val, y=y_start)
    pendown()
    goto(x=x_val, y=y_end)
    
done()		    # Complete drawings



# play_sound.py    31Mar2023  crs, extract to move to speaker_control

import datetime
from select_trace import SlTrace
try:
    from sinewave_numpy import SineWaveNumPy
    has_sinewave = True 
except:
    has_sinewave = False 

class PlaySound:
    """ particular sound to be played
    """

    @classmethod
    def pskey(cls, pitch, volume):
        """ Get unique resource key for possible reuse of SineWaveNumPy still needed???
        :pitch: sound pitch
        :volume: sound volume decibel if  not tuple then (volume,volume)
        :returns: unique resource key
        """
        if type(volume) != tuple:
            volume = (volume,volume)
        return (pitch,volume)
    
    def __init__(self, psc, pitch, dur=None,
                 volume=None, pskey=None):
        """ Setup sound to play
        :psc: controler
        :pitch: pitch to play
        :dur: duration of play (msec) default:psc.dur_base
        :volume: stereo volume to play (decibels) default:psc.volume_base
        :pskey: unique key  for PlaySound resource default: see pskey()
        """
        self.psid = psc.new_id() # Store unique id
        
        if volume is None:
            volume = psc.volume 
        if dur is None:
            dur = psc.dur
        self.psc = psc
        self.pitch = pitch
        self.time_play_start = None      # Play start time, if any 
        self.time_play_stop = None       # Play stop time, if any
        if dur is None:
            dur = psc.dur
        self.dur = dur
        if not isinstance(dur, (int, float)):
            SlTrace.lg(f"bad type for dur {self.dur}:{type(self.dur)}")
            dur  = .2
        if type(volume) != tuple:
            volume = (volume,volume)
        self.volume = volume
        volume_left,volume_right = volume
        self.waveform_stereo = SineWaveNumPy(pitch=pitch, decibels_left=volume_left,
                                      decibels_right=volume_right)
        if pskey is None:
            pskey = self.pskey(pitch,volume)
        self.pskey = pskey
        SlTrace.lg(f"New : {self}", "sound")

    def __str__(self):
        """ String instance for pretty printing
        """
        st = f"PlaySound[{self.psid}]:"
        st += f" pitch:{self.pitch}" 
        st += f" dur:{self.dur}"
        st += f" volume:{self.volume}"
        st += f" pskey:{self.pskey}"
        return st 

    def update_id(self):
        """ Update id for reuse
        """
        PlaySound.g_psid += 1
        self.psid = self.g_psid
        SlTrace.lg(f"Reusing PlaySound: {self}")
          
    def play(self, dur=None, dly=None):
        """ Play/replay sound for duration
        :dur: play duration default: initial duration (msec)
        :dly: delay (silence) before sound
        """
        self.update()
        if dur is None:     # Setup for delayed
            dur = self.dur
        if not isinstance(dur, (int, float)):
            self.waveform_stereo.stop()
            SlTrace.lg(f"bad type for dur {self.dur}:{type(self.dur)}")
            pass
        self.dur = dur
        self.is_active = True
        if dly is not None and dly != 0:
            SlTrace.lg(f"delaying play: dly:{dly} {self}", "sound")
            self.psc.awin.canvas.after(dly, self.play_sound)
            self.update()
            return
        
        self.play_it()
        
    def play_it(self):
        """ direct play
        """
        SlTrace.lg(f"play( ps:{self}", "sound")
        self.time_play_start = datetime.datetime.now()
        self.time_play_stop = None
        self.waveform_stereo.play()
        if not isinstance(self.dur, (int, float)):
            SlTrace.lg(f"bad type for self.dur {self.dur}:{type(self.dur)}")
            self.waveform_stereo.stop()
            return
        
        
        self.update()        
        
    def play_sound(self):
        """ Play sound now, possibly delayed
        """
        self.play_it()
         
    def play_stop(self):
        """ Stop playing - called after duration of play
        """
        if not self.is_active:
            return 
        self.time_play_stop = datetime.datetime.now()
        play_dur = (self.time_play_stop - self.time_play_start).total_seconds()
        SlTrace.lg(f"play_stop play_dur:{play_dur:.3} ps:{self}", "sound")
        self.waveform_stereo.stop()
        self.psc.free_ps(self)     # Release resource - put on available list

    def update(self):
        self.psc.update()
    
    def update_pending(self):
        """ Update pending events
        """
        self.psc.update_pending()



# java_properties.py    20Sep2018
"""
Java properties file processing
Simple single name.name.... = value
"""
import os
import sys
import copy
import re

from crs_funs import dot_sorted

class JavaProperties:
    @classmethod
    def snapshot_properties(cls, sn=None,
                                req=None, req_not=False,
                                req_match=None,
                                req_match_not=False):
        """ Properties snapshot, selected by req string and or req_match
        :sn: properties state
            default: REQUIRED
        :req: only keys which have this string
        :req_not: only keys that don't have this string
        :req_match: only keys which match this regular expression
        :req_match_not: only keys which don't match this regex
                
        :returns: (JavaProperties(only in sn1),
                  JavaProperties(sn1, sn1,sn2 differ),
                  JavaProperties(sn2, sn1,sn2 differ),
                  JavaProperties (only in sn2)) tuple
        """
        if req is None and req_match is None:
            return sn.copy()
        
        jp_in_sn = JavaProperties()
        for prop_key in sn.getPropKeys():
            if cls.is_our_key(prop_key, req=req, req_not=req_not, req_match=req_match, req_match_not=req_match_not):
                jp_in_sn.setProperty(prop_key, sn.getProperty(prop_key, None))
        return jp_in_sn



    @classmethod
    def is_our_key(cls, prop_key, req=None, req_not=False, req_match=None, req_match_not=False):
        """ Test if this is our key
        """
        if req is not None:
            if prop_key.find(req) >= 0:
                if req_not:
                    return False                # string found but not wanted - skip it
            else:
                return False                    # Required string not found - skip it
    
        if req_match is not None:
            if re.match(req_match, prop_key) is not None:
                if req_match_not:
                    return False                # string found but not wanted - skip it
            else:
                if not req_match_not:
                    return False                    # Required string not found - skip it
        return True                                 # Our key

    
    def __init__(self, filename=None):
        self.props = {}
        self.filename = None
        if filename is not None:
            self.props = self.load(filename)

    def copy(self):
        """ Copy current properties data snapshot
        elemebts copied to protect from changes to source
        """
        new_copy = JavaProperties()
        for key in self.props:
            new_copy.props[key] = copy.copy(self.props[key])
        return new_copy
                
    def load(self, filename):
        sep = "="
        if "." not in(filename):
            filename += ".properties"   # Default extension
        if not os.path.isabs(filename):
            filename = os.path.abspath(filename)
        self.filename = filename
        if not os.path.exists(filename):
            with open(filename, "w") as fout:
                abs_filename = os.path.abspath(filename)
                print("Creatng empty file %s" % abs_filename)
                print("# %s Empty - created" % filename, file=fout)
                fout.close()
        with open(filename) as fin:
            props = {}
            for line in fin:
                if "#" in line:
                    ic = line.index("#")
                    line = line[ic:]
                if sep not in line:
                    continue
                            # Find the name and value by splitting the string
                name, value = line.split(sep, 1)
    
                # Assign key value pair to dict
                # strip() removes white space from the ends of strings
                props[name.strip()] = value.strip()
        return props

    def getPropKeys(self, startswith=None):
        """ Get list of keys
        :startswith: starting string
            default: All keys
        :returns: keys sorted dot wise (dot_sorted)
        """
        if startswith is None:
            return dot_sorted(self.props.keys())
        
        keys = []
        for key in dot_sorted(self.props.keys()):
            if key.startswith(startswith):
                keys.append(key)
        return keys

    def getPropTree(self, startswith=None):
        """ Get tree based on doted keys
           head.next_sections.....
           :startswith: restrict tree to those keys that start with this string
                           default: all keys
        :returns: a dictionary tree by segment values each containing a dictionary of 
                the subsequent sections, till the bottom leaf.
                a.b.c, a.c, a.d, a.b.c2, b.c2 => {a, b}  a: {b,c,}  b: {c, c2}
                each value will be the 
        """
        tree = {}
        keys = getPropKeys(startswith=startswith)
        for key in keys:
            self.add_key_to_tree(key)
        ### TBD
            
        
        
    def getProperty(self, key, default):
        """ Get property, returning default if none
        """
        try:
            value = self.props[key]
        except:
            value = default
        return value
    
        
    def get_properties(self):
        """ Return dictionary of keys and value text strings
        """
        return self.props


    def hasProp(self, key):
        """ Check if property is already present
        :key: property key
        :returns: True iff property is present
        """
        if key in self.props:
            return True
        
        return False

    def is_empty(self):
        """ Check if no properties
        """
        if len(self.props) == 0:
            return True
        
        return False

    def deleteProperty(self, key):
        """ Delete property(key)
        """
        if self.hasProp(key):
            del self.props[key]


    def setProperty(self, key, value):
        """ Set property(key) value
        Converts value to string before storing
        """
        self.props[key] = str(value)


    def store(self, fp, title=None, keys=None, list_props=True):
        """ Store properties file
        :fp: Output file
        :title: Optional comment line
        :keys: list of property keys to store
                default: self.props
        :list_props: List properties on stdout
        """
        if keys is None:
            keys = self.props.keys()
        try:
            if title is not None:
                print("# %s" % title, file=fp)
            for key in dot_sorted(keys):
                prop_str = f"{key}={self.props[key]}"
                print(prop_str, file=fp)
                if list_props:
                    print(prop_str)
            fp.close()
        except IOError as ioex:
            print("Error in storing Properties file %s" % title)
            print("errno: %d" % ioex.errno)
            print("err code: %d" % ioex.errorcode[ioex.errno])
            print("err message: %s" % os.strerror(ioex.errno))

    
    def get_path(self):
        """ Get full path to properties file
        """
        return self.filename        

    
if __name__ == "__main__":
    propfile = "test"
    pr = JavaProperties(propfile)
    dp = pr.get_properties()
    
    for key in dp:
        print("dp[%s] = %s" % (key, dp[key]))
    
    pr2 = JavaProperties(propfile)
    pk = "new_key"
    default_val = "check_val"
    pval = pr2.getProperty(pk, default_val)
    if pval == default_val:
        print("Got expected default %s value %s" % (pk, pval))
    else:
        print("Got %s NOT the EXPECTED %s" % (pval, default_val))
        print("Quitting")
        sys.exit(1)
        
    new_value = "New VALUE"   
    pr2.setProperty(pk, new_value)
    
    propfile2 = "test2.properties"
    abs_propfile2 = os.path.abspath(propfile2)
    print("\nWriting new properties file %s" % abs_propfile2)    
    with open(propfile2, "w") as p2f:
        pr2.store(p2f, "# %s properties file 2" % propfile2)
        p2f.close()

    print("Loading new properties file %s" % propfile2)    
    pr2b = JavaProperties(propfile2)
    pk2 = pk
    pval2 = pr2b.getProperty(pk2, None)
    if pval2 is None:
        print("Property %s not found" % pk2)
        sys.exit(1)
        
    print("pval=%s" % pval2)
    print("new_value=%s" % new_value)
    if pval2 != pval2:
        print("pval2 Not Equal to pval")
        print("Property key:%s: %s was NOT the EXPECTED %s"
               % (pk2, pval2, pval))
        print("Quitting")
        sys.exit
    print(f"deleting key={pk}")
    if pr2b.hasProp(pk):
        print(f"Before: property {pk} is here")
    pr2b.deleteProperty(pk)
    if not pr2b.hasProp(pk):
        print(f"After: property {pk} is gone")
    if pr2b.hasProp(pk):
        print(f"property {pk} is NOT gone")
    print("End of Test")        
        